1.35a psp bugfix release
[libpicofe.git] / psp / emu.c
index e72ea7d..c63c00b 100644 (file)
--- a/psp/emu.c
+++ b/psp/emu.c
@@ -1,3 +1,8 @@
+// (c) Copyright 2007 notaz, All rights reserved.
+// Free for non-commercial use.
+
+// For commercial use, separate licencing terms must be obtained.
+
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/syslimits.h> // PATH_MAX
@@ -11,6 +16,7 @@
 #include "psp.h"
 #include "menu.h"
 #include "emu.h"
+#include "mp3.h"
 #include "../common/emu.h"
 #include "../common/lprintf.h"
 #include "../../Pico/PicoInt.h"
@@ -48,16 +54,16 @@ void emu_getMainDir(char *dst, int len)
        if (len > 0) *dst = 0;
 }
 
-static void osd_text(int x, const char *text, int is_active)
+static void osd_text(int x, const char *text, int is_active, int clear_all)
 {
        unsigned short *screen = is_active ? psp_video_get_active_fb() : psp_screen;
-       int len = strlen(text) * 8 / 2;
+       int len = clear_all ? (480 / 2) : (strlen(text) * 8 / 2);
        int *p, h;
        void *tmp;
        for (h = 0; h < 8; h++) {
                p = (int *) (screen+x+512*(264+h));
                p = (int *) ((int)p & ~3); // align
-               memset32(p, 0, len);
+               memset32_uncached(p, 0, len);
        }
        if (is_active) { tmp = psp_screen; psp_screen = screen; } // nasty pointer tricks
        emu_textOut16(x, 264, text);
@@ -66,7 +72,7 @@ static void osd_text(int x, const char *text, int is_active)
 
 void emu_msg_cb(const char *msg)
 {
-       osd_text(4, msg, 1);
+       osd_text(4, msg, 1, 1);
        noticeMsgTime = sceKernelGetSystemTimeLow() - 2000000;
 
        /* assumption: emu_msg_cb gets called only when something slow is about to happen */
@@ -126,8 +132,8 @@ void emu_setDefaultConfig(void)
 {
        memset(&currentConfig, 0, sizeof(currentConfig));
        currentConfig.lastRomFile[0] = 0;
-       currentConfig.EmuOpt  = 0x1f | 0x680; // | confirm_save, cd_leds, 16bit rend
-       currentConfig.PicoOpt = 0x0f | 0xc00; // | cd_pcm, cd_cdda
+       currentConfig.EmuOpt  = 0x1d | 0x680;  // | confirm_save, cd_leds, acc rend
+       currentConfig.PicoOpt = 0x0f | 0x1c00; // | gfx_cd, cd_pcm, cd_cdda
        currentConfig.PsndRate = 22050;
        currentConfig.PicoRegion = 0; // auto
        currentConfig.PicoAutoRgnOrder = 0x184; // US, EU, JP
@@ -145,9 +151,15 @@ void emu_setDefaultConfig(void)
        currentConfig.KeyBinds[12] = 1<<26; // switch rnd
        currentConfig.KeyBinds[ 8] = 1<<27; // save state
        currentConfig.KeyBinds[ 9] = 1<<28; // load state
-       currentConfig.PicoCDBuffers = 0;
-       currentConfig.scaling = 1; // bilinear filtering for psp
-       currentConfig.scale = currentConfig.hscale32 = currentConfig.hscale40 = 1.0;
+       currentConfig.KeyBinds[28] = 1<<0; // num "buttons"
+       currentConfig.KeyBinds[30] = 1<<1;
+       currentConfig.KeyBinds[31] = 1<<2;
+       currentConfig.KeyBinds[29] = 1<<3;
+       currentConfig.PicoCDBuffers = 64;
+       currentConfig.scaling = 1;     // bilinear filtering for psp
+       currentConfig.scale = 1.20;    // fullscreen
+       currentConfig.hscale40 = 1.25;
+       currentConfig.hscale32 = 1.56;
 }
 
 
@@ -166,7 +178,7 @@ static int fbimg_offs = 0;
 
 static void set_scaling_params(void)
 {
-       int src_width, fbimg_width, fbimg_height, fbimg_xoffs, fbimg_yoffs;
+       int src_width, fbimg_width, fbimg_height, fbimg_xoffs, fbimg_yoffs, border_hack = 0;
        g_vertices[0].x = g_vertices[0].y =
        g_vertices[0].z = g_vertices[1].z = 0;
 
@@ -179,9 +191,12 @@ static void set_scaling_params(void)
                src_width = 256;
        }
 
+       if (fbimg_width  & 1) fbimg_width++;  // make even
+       if (fbimg_height & 1) fbimg_height++;
+
        if (fbimg_width >= 480) {
                g_vertices[0].u = (fbimg_width-480)/2;
-               g_vertices[1].u = src_width - (fbimg_width-480)/2;
+               g_vertices[1].u = src_width - (fbimg_width-480)/2 - 1;
                fbimg_width = 480;
                fbimg_xoffs = 0;
        } else {
@@ -189,6 +204,7 @@ static void set_scaling_params(void)
                g_vertices[1].u = src_width;
                fbimg_xoffs = 240 - fbimg_width/2;
        }
+       if (fbimg_width > 320 && fbimg_width <= 480) border_hack = 1;
 
        if (fbimg_height >= 272) {
                g_vertices[0].v = (fbimg_height-272)/2;
@@ -205,6 +221,12 @@ static void set_scaling_params(void)
        g_vertices[1].y = fbimg_height;
        if (fbimg_xoffs < 0) fbimg_xoffs = 0;
        if (fbimg_yoffs < 0) fbimg_yoffs = 0;
+       if (border_hack) {
+               g_vertices[0].u++;
+               g_vertices[0].x++;
+               g_vertices[1].u--;
+               g_vertices[1].x--;
+       }
        fbimg_offs = (fbimg_yoffs*512 + fbimg_xoffs) * 2; // dst is always 16bit
 
        /*
@@ -353,21 +375,19 @@ static void blitscreen_clut(void)
 
 static void cd_leds(void)
 {
-       static int old_reg = 0;
-       unsigned int col_g, col_r, *p;
+       unsigned int reg, col_g, col_r, *p;
 
-       if (!((Pico_mcd->s68k_regs[0] ^ old_reg) & 3)) return; // no change
-       old_reg = Pico_mcd->s68k_regs[0];
+       reg = Pico_mcd->s68k_regs[0];
 
        p = (unsigned int *)((short *)psp_screen + 512*2+4+2);
-       col_g = (old_reg & 2) ? 0x06000600 : 0;
-       col_r = (old_reg & 1) ? 0x00180018 : 0;
+       col_g = (reg & 2) ? 0x06000600 : 0;
+       col_r = (reg & 1) ? 0x00180018 : 0;
        *p++ = col_g; *p++ = col_g; p+=2; *p++ = col_r; *p++ = col_r; p += 512/2 - 12/2;
        *p++ = col_g; *p++ = col_g; p+=2; *p++ = col_r; *p++ = col_r; p += 512/2 - 12/2;
        *p++ = col_g; *p++ = col_g; p+=2; *p++ = col_r; *p++ = col_r;
 }
 
-
+#if 0
 static void dbg_text(void)
 {
        int *p, h, len;
@@ -378,11 +398,11 @@ static void dbg_text(void)
        for (h = 0; h < 8; h++) {
                p = (int *) ((unsigned short *) psp_screen+2+512*(256+h));
                p = (int *) ((int)p & ~3); // align
-               memset32(p, 0, len);
+               memset32_uncached(p, 0, len);
        }
        emu_textOut16(2, 256, text);
 }
-
+#endif
 
 /* called after rendering is done, but frame emulation is not finished */
 void blit1(void)
@@ -407,11 +427,11 @@ static void blit2(const char *fps, const char *notice, int lagging_behind)
        int vsync = 0, emu_opt = currentConfig.EmuOpt;
 
        if (notice || (emu_opt & 2)) {
-               if (notice)      osd_text(4, notice, 0);
-               if (emu_opt & 2) osd_text(OSD_FPS_X, fps, 0);
+               if (notice)      osd_text(4, notice, 0, 0);
+               if (emu_opt & 2) osd_text(OSD_FPS_X, fps, 0, 0);
        }
 
-       dbg_text();
+       //dbg_text();
 
        if ((emu_opt & 0x400) && (PicoMCD & 1))
                cd_leds();
@@ -427,15 +447,15 @@ static void blit2(const char *fps, const char *notice, int lagging_behind)
 static void clearArea(int full)
 {
        if (full) {
-               memset32(psp_screen, 0, 512*272*2/4);
+               memset32_uncached(psp_screen, 0, 512*272*2/4);
                psp_video_flip(0);
-               memset32(psp_screen, 0, 512*272*2/4);
+               memset32_uncached(psp_screen, 0, 512*272*2/4);
                memset32(VRAM_CACHED_STUFF, 0xe0e0e0e0, 512*240/4);
                memset32((int *)VRAM_CACHED_STUFF+512*240/4, 0, 512*240*2/4);
        } else {
                void *fb = psp_video_get_active_fb();
-               memset32((int *)((char *)psp_screen + 512*264*2), 0, 512*8*2/4);
-               memset32((int *)((char *)fb         + 512*264*2), 0, 512*8*2/4);
+               memset32_uncached((int *)((char *)psp_screen + 512*264*2), 0, 512*8*2/4);
+               memset32_uncached((int *)((char *)fb         + 512*264*2), 0, 512*8*2/4);
        }
 }
 
@@ -473,7 +493,7 @@ static void vidResetMode(void)
 /* sound stuff */
 #define SOUND_BLOCK_SIZE_NTSC (1470*2) // 1024 // 1152
 #define SOUND_BLOCK_SIZE_PAL  (1764*2)
-#define SOUND_BLOCK_COUNT    4
+#define SOUND_BLOCK_COUNT    8
 
 static short __attribute__((aligned(4))) sndBuffer[SOUND_BLOCK_SIZE_PAL*SOUND_BLOCK_COUNT + 44100/50*2];
 static short *snd_playptr = NULL, *sndBuffer_endptr = NULL;
@@ -485,7 +505,7 @@ static void writeSound(int len);
 
 static int sound_thread(SceSize args, void *argp)
 {
-       int ret;
+       int ret = 0;
 
        lprintf("sthr: started, priority %i\n", sceKernelGetThreadCurrentPriority());
 
@@ -496,11 +516,11 @@ static int sound_thread(SceSize args, void *argp)
                        //lprintf("sthr: wait... (%i)\n", samples_made - samples_done);
                        while (samples_made - samples_done <= samples_block*2 && !sound_thread_exit)
                                ret = sceKernelWaitSema(sound_sem, 1, 0);
-                       //lprintf("sthr: sceKernelWaitSema: %i\n", ret);
+                       if (ret < 0) lprintf("sthr: sceKernelWaitSema: %i\n", ret);
                        continue;
                }
 
-               //lprintf("sthr: got data: %i\n", samples_made - samples_done);
+               // lprintf("sthr: got data: %i\n", samples_made - samples_done);
 
                ret = sceAudio_E0727056(PSP_AUDIO_VOLUME_MAX, snd_playptr);
 
@@ -508,12 +528,13 @@ static int sound_thread(SceSize args, void *argp)
                snd_playptr  += samples_block;
                if (snd_playptr >= sndBuffer_endptr)
                        snd_playptr = sndBuffer;
-               if (ret)
-                       lprintf("sthr: outf: %i; pos %i/%i\n", ret, samples_done, samples_made);
+               // 1.5 kernel returns 0, newer ones return # of samples queued
+               if (ret < 0)
+                       lprintf("sthr: sceAudio_E0727056: %08x; pos %i/%i\n", ret, samples_done, samples_made);
 
                // shouln't happen, but just in case
                if (samples_made - samples_done >= samples_block*3) {
-                       //lprintf("block skip (%i)\n", samples_made - samples_done);
+                       lprintf("sthr: block skip (%i)\n", samples_made - samples_done);
                        samples_done += samples_block; // skip
                        snd_playptr  += samples_block;
                }
@@ -528,6 +549,7 @@ static int sound_thread(SceSize args, void *argp)
 static void sound_init(void)
 {
        SceUID thid;
+       int ret;
 
        sound_sem = sceKernelCreateSema("sndsem", 0, 0, 1, NULL);
        if (sound_sem < 0) lprintf("sceKernelCreateSema() failed: %i\n", sound_sem);
@@ -538,7 +560,8 @@ static void sound_init(void)
        thid = sceKernelCreateThread("sndthread", sound_thread, 0x12, 0x10000, 0, NULL);
        if (thid >= 0)
        {
-               sceKernelStartThread(thid, 0, 0);
+               ret = sceKernelStartThread(thid, 0, 0);
+               if (ret < 0) lprintf("sound_init: sceKernelStartThread returned %08x\n", ret);
        }
        else
                lprintf("sceKernelCreateThread failed: %i\n", thid);
@@ -563,8 +586,8 @@ static void sound_prepare(void)
        lprintf("starting audio: %i, len: %i, stereo: %i, pal: %i, block samples: %i\n",
                        PsndRate, PsndLen, stereo, Pico.m.pal, samples_block);
 
-       while (sceAudioOutput2GetRestSample() > 0) psp_msleep(100);
-       sceAudio_5C37C0AE();
+       // while (sceAudioOutput2GetRestSample() > 0) psp_msleep(100);
+       // sceAudio_5C37C0AE();
        ret = sceAudio_38553111(samples_block/2, PsndRate, 2); // seems to not need that stupid 64byte alignment
        if (ret < 0) {
                lprintf("sceAudio_38553111() failed: %i\n", ret);
@@ -585,8 +608,19 @@ static void sound_prepare(void)
 
 static void sound_end(void)
 {
+       int i;
+       if (samples_done == 0)
+       {
+               // if no data is written between sceAudio_38553111 and sceAudio_5C37C0AE calls,
+               // we get a deadlock on next sceAudio_38553111 call
+               // so this is yet another workaround:
+               memset32((int *)(void *)sndBuffer, 0, samples_block*4/4);
+               samples_made = samples_block * 3;
+               sceKernelSignalSema(sound_sem, 1);
+       }
+       sceKernelDelayThread(100*1000);
        samples_made = samples_done = 0;
-       while (sceAudioOutput2GetRestSample() > 0)
+       for (i = 0; sceAudioOutput2GetRestSample() > 0 && i < 16; i++)
                psp_msleep(100);
        sceAudio_5C37C0AE();
 }
@@ -611,6 +645,7 @@ static void writeSound(int len)
                lprintf("mov\n");
        }
        else*/
+       if (PsndOut > sndBuffer_endptr) lprintf("snd oflow %i!\n", PsndOut - sndBuffer_endptr);
        if (PsndOut >= sndBuffer_endptr)
                PsndOut = sndBuffer;
 
@@ -619,7 +654,7 @@ static void writeSound(int len)
        if (samples_made - samples_done > samples_block*2) {
                // lprintf("signal, %i/%i\n", samples_done, samples_made);
                ret = sceKernelSignalSema(sound_sem, 1);
-               // lprintf("signal ret %i\n", ret);
+               //if (ret < 0) lprintf("snd signal ret %08x\n", ret);
        }
 }
 
@@ -643,7 +678,7 @@ void emu_forcedFrame(void)
        vidResetMode();
        memset32(VRAM_CACHED_STUFF, 0xe0e0e0e0, 512*8/4); // borders
        memset32((int *)VRAM_CACHED_STUFF + 512*232/4, 0xe0e0e0e0, 512*8/4);
-       memset32((int *)psp_screen + 512*264*2/4, 0, 512*8*2/4);
+       memset32_uncached((int *)psp_screen + 512*264*2/4, 0, 512*8*2/4);
 
        PicoDrawSetColorFormat(-1);
        PicoScan = EmuScanSlow;
@@ -680,7 +715,7 @@ static void RunEvents(unsigned int which)
 
                if (do_it)
                {
-                       osd_text(4, (which & 0x1000) ? "LOADING GAME" : "SAVING GAME", 1);
+                       osd_text(4, (which & 0x1000) ? "LOADING GAME" : "SAVING GAME", 1, 0);
                        PicoStateProgressCB = emu_msg_cb;
                        emu_SaveLoadGame((which & 0x1000) >> 12, 0);
                        PicoStateProgressCB = NULL;
@@ -798,23 +833,22 @@ static void find_combos(void)
                if (act == 16 || act == 17) continue; // player2 flag
                if (act > 17)
                {
-                       for (u = 0; u < 32; u++)
+                       for (u = 0; u < 28; u++) // 28 because nub can't produce combos
                                if (currentConfig.KeyBinds[u] & (1 << act)) keyc++;
                }
                else
                {
-                       for (u = 0; u < 32; u++)
+                       for (u = 0; u < 28; u++)
                                if ((currentConfig.KeyBinds[u] & 0x30000) == 0 && // pl. 1
                                        (currentConfig.KeyBinds[u] & (1 << act))) keyc++;
-                       for (u = 0; u < 32; u++)
+                       for (u = 0; u < 28; u++)
                                if ((currentConfig.KeyBinds[u] & 0x30000) == 1 && // pl. 2
                                        (currentConfig.KeyBinds[u] & (1 << act))) keyc2++;
-                       if (keyc2 > keyc) keyc = keyc2;
                }
-               if (keyc > 1)
+               if (keyc > 1 || keyc2 > 1)
                {
                        // loop again and mark those keys and actions as combo
-                       for (u = 0; u < 32; u++)
+                       for (u = 0; u < 28; u++)
                        {
                                if (currentConfig.KeyBinds[u] & (1 << act)) {
                                        combo_keys |= 1 << u;
@@ -839,6 +873,7 @@ static void simpleWait(unsigned int until)
 
 void emu_Loop(void)
 {
+       static int mp3_init_done = 0;
        char fpsbuff[24]; // fps count c string
        unsigned int tval, tval_prev = 0, tval_thissec = 0; // timing
        int frames_done = 0, frames_shown = 0, oldmodes = 0;
@@ -868,8 +903,16 @@ void emu_Loop(void)
        target_frametime = Pico.m.pal ? (1000000<<8)/50 : (1000000<<8)/60+1;
        reset_timing = 1;
 
-       // prepare CD buffer
-       if (PicoMCD & 1) PicoCDBufferInit();
+       if (PicoMCD & 1) {
+               // prepare CD buffer
+               PicoCDBufferInit();
+               // mp3...
+               if (!mp3_init_done) {
+                       i = mp3_init();
+                       mp3_init_done = 1;
+                       if (i) { engineState = PGS_Menu; return; }
+               }
+       }
 
        // prepare sound stuff
        PsndOut = NULL;
@@ -947,7 +990,7 @@ void emu_Loop(void)
                        for (i = 0; i < currentConfig.Frameskip; i++) {
                                updateKeys();
                                SkipFrame(); frames_done++;
-                               if (PsndOut) { // do framelimitting if sound is enabled
+                               if (!(currentConfig.EmuOpt&0x40000)) { // do framelimitting if needed
                                        int tval_diff;
                                        tval = sceKernelGetSystemTimeLow();
                                        tval_diff = (int)(tval - tval_thissec) << 8;
@@ -992,7 +1035,7 @@ void emu_Loop(void)
 
                if (currentConfig.Frameskip < 0 && tval_diff - lim_time >= (300000<<8)) // slowdown detection
                        reset_timing = 1;
-               else if (PsndOut != NULL || currentConfig.Frameskip < 0)
+               else if (!(currentConfig.EmuOpt&0x40000) || currentConfig.Frameskip < 0)
                {
                        // sleep if we are still too fast
                        if (tval_diff < lim_time)
@@ -1021,7 +1064,7 @@ void emu_Loop(void)
        }
 
        // clear fps counters and stuff
-       memset32((int *)psp_video_get_active_fb() + 512*264*2/4, 0, 512*8*2/4);
+       memset32_uncached((int *)psp_video_get_active_fb() + 512*264*2/4, 0, 512*8*2/4);
 }